import time

import pytest

import parsl
from parsl import ThreadPoolExecutor, python_app
from parsl.config import Config
from parsl.data_provider.files import File
from parsl.data_provider.staging import Staging


@python_app
def observe_input_local_path(f):
    """Returns the local_path that is seen by the app,
    so that the test suite can also see that path.
    """
    return f.local_path


@python_app
def wait_and_create(outputs=[]):
    # for test purposes, this doesn't actually need to create the output
    # file as nothing ever touches file content - the test only deals with
    # names and Futures.
    time.sleep(10)


class SP2(Staging):
    def can_stage_in(self, file):
        return file.scheme == 'sp2'

    def stage_in(self, dm, executor, file, parent_fut):
        file.local_path = "./test1.tmp"
        return None  # specify no tasks inserted in graph so parsl will preserve stageout dependency

    def can_stage_out(self, file):
        return file.scheme == 'sp2'

    def stage_out(self, dm, executor, file, app_fu):
        file.local_path = "./test1.tmp"
        return None  # no tasks inserted so parsl should use app completion for DataFuture completion


@pytest.mark.local
def test_1316_local_path_on_execution_side_sp2():
    """This test demonstrates the ability of a StagingProvider to set the
    local_path of a File on the execution side, but that the change does not
    modify the local_path of the corresponding submit side File, even when
    running in a single python process.
    """

    config = Config(executors=[ThreadPoolExecutor(storage_access=[SP2()])])

    file = File("sp2://test")

    parsl.load(config)
    p = observe_input_local_path(file).result()

    assert p == "./test1.tmp", "File object on the execution side gets the local_path set by the staging provider"

    assert not file.local_path, "The local_path on the submit side should not be set"

    parsl.dfk().cleanup()


@pytest.mark.local
def test_1316_local_path_setting_preserves_dependency_sp2():
    config = Config(executors=[ThreadPoolExecutor(storage_access=[SP2()])])

    file = File("sp2://test")

    parsl.load(config)

    wc_app_future = wait_and_create(outputs=[file])
    data_future = wc_app_future.outputs[0]

    p = observe_input_local_path(data_future).result()

    assert wc_app_future.done(), "wait_and_create should finish before observe_input_local_path finishes"

    assert p == "./test1.tmp", "File object on the execution side gets the local_path set by the staging provider"

    assert not file.local_path, "The local_path on the submit side should not be set"

    parsl.dfk().cleanup()



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import GridEngineProvider

from .user_opts import user_opts


def fresh_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                label='cc_in2p3_htex',
                max_workers_per_node=1,
                encrypted=True,
                provider=GridEngineProvider(
                    nodes_per_block=2,
                    init_blocks=2,
                    max_blocks=2,
                    walltime="00:20:00",
                    scheduler_options=user_opts['cc_in2p3']['scheduler_options'],
                    worker_init=user_opts['cc_in2p3']['worker_init'],
                ),
            )
        ],
    )


config = fresh_config()



from abc import ABCMeta, abstractmethod
from multiprocessing.queues import Queue


class MonitoringRadioReceiver(metaclass=ABCMeta):
    @abstractmethod
    def shutdown(self) -> None:
        pass


class MonitoringRadioSender(metaclass=ABCMeta):
    @abstractmethod
    def send(self, message: object) -> None:
        pass


class RadioConfig(metaclass=ABCMeta):
    """Base class for radio plugin configuration.

    This provides the configuration for a particular way of sending monitoring
    messages from a source of monitoring messages into the submit side
    monitoring database.

    This uses staged initialization like lots of Parsl configuration, but in
    a slightly different form.

    A RadioConfig object must be pickleable, because it will be sent to remote
    workers to configure senders. The MonitoringRadioSender and
    MonitoringRadioReceiver objects do not need to be pickleable (and often
    will not be - for example, when they hold references to other processes).

    The RadioConfig object will be used by Parsl in this sequence:

    * A user creates a RadioConfig object from the appropriate subclass for
      radio mechanism they want to use, and specifies it as part of their
      executor configuration.

    Methods on the RadioConfig will then be invoked by Parsl like this:

     * one create_receiver call, on the submit side
        - this call can modify the state of radioconfig to contain information
          about how a sender can connect back to the receiver. for example,
          after binding to a particular port, can store that port so that the
          sender knows which port to connect to.

     * Possibly many serializations to get the RadioConfig to remote workers

     * Many (0 or more) create_sender calls, possibly on remote workers, to
       create the sending side of the radio (MonitoringRadioSender instances)

     * Those senders are used to send messages

     * At executor shutdown, the receiver is shut down.

    This object cannot be re-used across parsl configurations - like many other
    pieces of parsl config it is single use in that respect.
    """

    @abstractmethod
    def create_receiver(self, *, run_dir: str, resource_msgs: Queue) -> MonitoringRadioReceiver:
        """Create a receiver for this RadioConfig, and update this RadioConfig
        with enough context to create senders.
        """
        pass

    @abstractmethod
    def create_sender(self) -> MonitoringRadioSender:
        """Create a sender to connect to the receiver created by an
        earlier call to create_receiver.
        """
        pass



"""Tests monitoring records app name under various decoration patterns.
"""

import logging
import os
import re
import time
from typing import Union

import pytest

import parsl
from parsl.config import Config
from parsl.data_provider.data_manager import default_staging
from parsl.data_provider.files import File
from parsl.data_provider.staging import Staging
from parsl.executors import HighThroughputExecutor
from parsl.monitoring import MonitoringHub
from parsl.providers import LocalProvider


def fresh_config(run_dir):
    return Config(
        run_dir=str(run_dir),
        executors=[
            HighThroughputExecutor(
                address="127.0.0.1",
                label="htex_Local",
                provider=LocalProvider(
                    init_blocks=1,
                    min_blocks=1,
                    max_blocks=1,
                )
            )
        ],
        strategy='simple',
        strategy_period=0.1,
        monitoring=MonitoringHub(),
    )


@parsl.python_app
def stdapp(stdout=None, stderr=None):
    pass


class ArbitraryPathLike(os.PathLike):
    def __init__(self, path: Union[str, bytes]) -> None:
        self.path = path

    def __fspath__(self) -> Union[str, bytes]:
        return self.path


class ArbitraryStaging(Staging):
    """This staging provider will not actually do any staging, but will
    accept arbitrary: scheme URLs. That's enough for this monitoring test
    which doesn't need any actual stage out action to happen.
    """
    def can_stage_out(self, file):
        return file.scheme == "arbitrary"


@pytest.mark.local
@pytest.mark.parametrize('stdx,expected_stdx',
                         [('hello.txt', 'hello.txt'),
                          (None, ''),
                          (('tuple.txt', 'w'), 'tuple.txt'),
                          (ArbitraryPathLike('pl.txt'), 'pl.txt'),
                          (ArbitraryPathLike(b'pl2.txt'), 'pl2.txt'),
                          ((ArbitraryPathLike('pl3.txt'), 'w'), 'pl3.txt'),
                          ((ArbitraryPathLike(b'pl4.txt'), 'w'), 'pl4.txt'),
                          (parsl.AUTO_LOGNAME,
                              lambda p:
                              isinstance(p, str) and
                              os.path.isabs(p) and
                              re.match("^.*/task_0000_stdapp\\.std...$", p)),
                          (File("arbitrary:abc123"), "arbitrary:abc123"),
                          (File("file:///tmp/pl5"), "file:///tmp/pl5"),
                          ])
@pytest.mark.parametrize('stream', ['stdout', 'stderr'])
def test_stdstream_to_monitoring(stdx, expected_stdx, stream, tmpd_cwd, caplog):
    """This tests that various forms of stdout/err specification are
       represented in monitoring correctly. The stderr and stdout codepaths
       are generally duplicated, rather than factorised, and so this test
       runs the same tests on both stdout and stderr.
    """

    # this is imported here rather than at module level because
    # it isn't available in a plain parsl install, so this module
    # would otherwise fail to import and break even a basic test
    # run.
    import sqlalchemy

    c = fresh_config(tmpd_cwd)
    c.monitoring.logging_endpoint = f"sqlite:///{tmpd_cwd}/monitoring.db"
    c.executors[0].storage_access = default_staging + [ArbitraryStaging()]

    with parsl.load(c):
        kwargs = {stream: stdx}
        stdapp(**kwargs).result()

    for record in caplog.records:
        assert record.levelno < logging.ERROR

    engine = sqlalchemy.create_engine(c.monitoring.logging_endpoint)
    with engine.begin() as connection:

        def count_rows(table: str):
            result = connection.execute(sqlalchemy.text(f"SELECT COUNT(*) FROM {table}"))
            (c, ) = result.first()
            return c

        # one workflow...
        assert count_rows("workflow") == 1

        # ... with one task ...
        assert count_rows("task") == 1

        # ... that was tried once ...
        assert count_rows("try") == 1

        # ... and has the expected name.
        result = connection.execute(sqlalchemy.text(f"SELECT task_{stream} FROM task"))
        (c, ) = result.first()

        if isinstance(expected_stdx, str):
            assert c == expected_stdx
        elif callable(expected_stdx):
            assert expected_stdx(c)
        else:
            raise RuntimeError("Bad expected_stdx value")

    for record in caplog.records:
        assert record.levelno < logging.ERROR



import logging
import os
import zipfile

import pytest

import parsl
from parsl.app.futures import DataFuture
from parsl.data_provider.files import File
from parsl.tests.configs.htex_local import fresh_config as local_config


@parsl.bash_app
def output_to_stds(*, stdout=parsl.AUTO_LOGNAME, stderr=parsl.AUTO_LOGNAME):
    return "echo hello ; echo goodbye >&2"


@pytest.mark.staging_required
def test_stdout_staging_file(tmpd_cwd, caplog):
    basename = str(tmpd_cwd) + "/stdout.txt"
    stdout_file = File("file://" + basename)

    app_future = output_to_stds(stdout=stdout_file)

    assert isinstance(app_future.stdout, DataFuture)
    app_future.stdout.result()

    assert os.path.exists(basename)

    for record in caplog.records:
        assert record.levelno < logging.ERROR


@pytest.mark.staging_required
def test_stdout_stderr_staging_zip(tmpd_cwd, caplog):
    zipfile_name = str(tmpd_cwd) + "/staging.zip"
    stdout_relative_path = "somewhere/test-out.txt"
    stdout_file = File("zip:" + zipfile_name + "/" + stdout_relative_path)

    stderr_relative_path = "somewhere/test-error.txt"
    stderr_file = File("zip:" + zipfile_name + "/" + stderr_relative_path)

    app_future = output_to_stds(stdout=stdout_file, stderr=stderr_file)

    assert isinstance(app_future.stdout, DataFuture)
    app_future.stdout.result()

    # check the file exists as soon as possible
    assert os.path.exists(zipfile_name)
    with zipfile.ZipFile(zipfile_name) as z:
        with z.open(stdout_relative_path) as f:
            assert f.readlines() == [b'hello\n']

    assert isinstance(app_future.stderr, DataFuture)
    app_future.stderr.result()
    with zipfile.ZipFile(zipfile_name) as z:
        with z.open(stderr_relative_path) as f:
            # The last line of stderr should be goodbye, but Parsl will write
            # other Parsl-specific into to stderr before that, so only assert
            # the behaviour of the final line.
            assert f.readlines()[-1] == b'goodbye\n'

    for record in caplog.records:
        assert record.levelno < logging.ERROR



import random
from unittest import mock

import pytest

from parsl.executors import GlobusComputeExecutor


@pytest.fixture
def mock_ex():
    # Not Parsl's job to test GC's Executor, although it
    # still needs to be importable for these test cases.
    from globus_compute_sdk import Executor

    yield mock.Mock(spec=Executor)


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_mock_spec(mock_ex):
    # a test of tests -- make sure we're using spec= in the mock
    with pytest.raises(AttributeError):
        mock_ex.aasdf()


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_label_default(mock_ex):
    gce = GlobusComputeExecutor(mock_ex)
    assert gce.label == type(gce).__name__, "Expect reasonable default label"


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_label(mock_ex, randomstring):
    exp_label = randomstring()
    gce = GlobusComputeExecutor(mock_ex, label=exp_label)
    assert gce.label == exp_label


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_resets_spec_after_submit(mock_ex, randomstring):
    submit_res = {randomstring(): "some submit res"}
    res = {"some": randomstring(), "spec": randomstring()}
    mock_ex.resource_specification = res
    mock_ex.user_endpoint_config = None
    gce = GlobusComputeExecutor(mock_ex)

    fn = mock.Mock()
    orig_res = mock_ex.resource_specification
    orig_uep = mock_ex.user_endpoint_config

    def mock_submit(*a, **k):
        assert mock_ex.resource_specification == submit_res, "Expect set for submission"
        assert mock_ex.user_endpoint_config is None
    mock_ex.submit.side_effect = mock_submit

    gce.submit(fn, resource_specification=submit_res)

    assert mock_ex.resource_specification == orig_res
    assert mock_ex.user_endpoint_config is orig_uep


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_resets_uep_after_submit(mock_ex, randomstring):
    uep_conf = randomstring()
    res = {"some": randomstring()}
    gce = GlobusComputeExecutor(mock_ex)

    fn = mock.Mock()
    orig_res = mock_ex.resource_specification
    orig_uep = mock_ex.user_endpoint_config

    def mock_submit(*a, **k):

        assert mock_ex.resource_specification == res, "Expect set for submission"
        assert mock_ex.user_endpoint_config == uep_conf, "Expect set for submission"
    mock_ex.submit.side_effect = mock_submit

    gce.submit(fn, resource_specification={"user_endpoint_config": uep_conf, **res})

    assert mock_ex.resource_specification == orig_res
    assert mock_ex.user_endpoint_config is orig_uep


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_happy_path(mock_ex, randomstring):
    mock_fn = mock.Mock()
    args = tuple(randomstring() for _ in range(random.randint(0, 3)))
    kwargs = {randomstring(): randomstring() for _ in range(random.randint(0, 3))}

    gce = GlobusComputeExecutor(mock_ex)
    gce.submit(mock_fn, {}, *args, **kwargs)

    assert mock_ex.submit.called, "Expect proxying of args to underlying executor"
    found_a, found_k = mock_ex.submit.call_args
    assert found_a[0] is mock_fn
    assert found_a[1:] == args
    assert found_k == kwargs


@pytest.mark.local
@pytest.mark.globus_compute
def test_gc_executor_shuts_down_asynchronously(mock_ex):
    gce = GlobusComputeExecutor(mock_ex)
    gce.shutdown()
    assert mock_ex.shutdown.called
    a, k = mock_ex.shutdown.call_args
    assert k["wait"] is False
    assert k["cancel_futures"] is True



from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer
from parsl.executors.threads import ThreadPoolExecutor

config = Config(
    executors=[
        ThreadPoolExecutor(
            label='local_threads_checkpoint_task_exit',
        )
    ],
    memoizer=BasicMemoizer(checkpoint_mode='task_exit')
)



import logging
import os
import pickle
import platform
import subprocess
import time

import psutil
import pytest
import zmq

import parsl.executors.high_throughput.zmq_pipes as zmq_pipes
from parsl.executors.high_throughput.executor import DEFAULT_INTERCHANGE_LAUNCH_CMD
from parsl.executors.high_throughput.manager_selector import RandomManagerSelector
from parsl.version import VERSION as PARSL_VERSION

P_ms = 10


@pytest.mark.local
def test_exit_with_bad_registration(tmpd_cwd, try_assert):
    """Test that the interchange exits when it receives a bad registration message.
    This complements parsl/tests/test_scaling/test_worker_interchange_bad_messages_3262.py
    which tests that the interchange is resistent to other forms of bad message.
    """

    outgoing_q = zmq_pipes.TasksOutgoing(
        "127.0.0.1", (49152, 65535), None
    )
    incoming_q = zmq_pipes.ResultsIncoming(
        "127.0.0.1", (49152, 65535), None
    )
    command_client = zmq_pipes.CommandClient(
        "127.0.0.1", (49152, 65535), None
    )

    interchange_config = {"client_address": "127.0.0.1",
                          "client_ports": (outgoing_q.port,
                                           incoming_q.port,
                                           command_client.port),
                          "interchange_address": "127.0.0.1",
                          "worker_port": None,
                          "worker_port_range": (50000, 60000),
                          "hub_address": None,
                          "hub_zmq_port": None,
                          "logdir": tmpd_cwd,
                          "heartbeat_threshold": 120,
                          "poll_period": P_ms,
                          "logging_level": logging.DEBUG,
                          "cert_dir": None,
                          "manager_selector": RandomManagerSelector(),
                          "run_id": "test",
                          "_check_python_mismatch": True,
                          }

    config_pickle = pickle.dumps(interchange_config)

    interchange_proc = subprocess.Popen(DEFAULT_INTERCHANGE_LAUNCH_CMD, stdin=subprocess.PIPE)
    stdin = interchange_proc.stdin
    assert stdin is not None, "Popen should have created an IO object (vs default None) because of PIPE mode"

    stdin.write(config_pickle)
    stdin.flush()
    stdin.close()

    # wait for interchange to be alive, by waiting for the command thread to become
    # responsive. if the interchange process didn't start enough to get the command
    # thread running, this will time out.

    worker_port = command_client.run("WORKER_BINDS", timeout_s=120)

    # now we'll assume that if the interchange command thread is responding,
    # then the worker polling code is also running and that the interchange has
    # started successfully.

    # send bad registration message as if from a new worker pool. The badness here
    # is that the Python version does not match the real Python version - which
    # unlike some other bad interchange messages, should cause the interchange
    # to shut down.

    msg = {'type': 'registration',
           'parsl_v': PARSL_VERSION,
           'python_v': "1.1.1",  # this is the bad bit
           'worker_count': 1,
           'uid': 'testuid',
           'block_id': 0,
           'start_time': time.time(),
           'prefetch_capacity': 0,
           'max_capacity': 1,
           'os': platform.system(),
           'hostname': platform.node(),
           'dir': os.getcwd(),
           'cpu_count': psutil.cpu_count(logical=False),
           'total_memory': psutil.virtual_memory().total,
           }

    # connect to worker port and send this message.

    context = zmq.Context()
    channel_timeout = 10000  # in milliseconds
    task_channel = context.socket(zmq.DEALER)
    task_channel.setsockopt(zmq.LINGER, 0)
    task_channel.setsockopt(zmq.IDENTITY, b'testid')

    task_channel.set_hwm(0)
    task_channel.setsockopt(zmq.SNDTIMEO, channel_timeout)
    task_channel.connect(f"tcp://127.0.0.1:{worker_port}")

    task_channel.send(pickle.dumps(msg))

    # check that the interchange exits within some reasonable time
    try_assert(lambda: interchange_proc.poll() is not None, "Interchange did not exit after killing watched client process", timeout_ms=5000)

    # See issue #3697 - ideally the interchange would exit cleanly, but it does not.
    # assert interchange_proc.poll() == 0, "Interchange exited with an error code, not 0"

    task_channel.close()
    context.term()



from globus_compute_sdk import Executor

from parsl.config import Config
from parsl.executors import GlobusComputeExecutor
from parsl.usage_tracking.levels import LEVEL_1

# Please start your own endpoint on perlmutter following instructions below to use this config:
# https://globus-compute.readthedocs.io/en/stable/endpoints/endpoint_examples.html#perlmutter-nersc
perlmutter_endpoint = 'YOUR_PERLMUTTER_ENDPOINT_UUID'

# Please start your own endpoint on expanse following instructions below to use this config:
# https://globus-compute.readthedocs.io/en/stable/endpoints/endpoint_examples.html#expanse-sdsc
expanse_endpoint = 'YOUR_EXPANSE_ENDPOINT_UUID'

config = Config(
    executors=[
        GlobusComputeExecutor(
            executor=Executor(endpoint_id=perlmutter_endpoint),
            label="Perlmutter",
        ),
        GlobusComputeExecutor(
            executor=Executor(endpoint_id=expanse_endpoint),
            label="Expanse",
        ),
    ],
    usage_tracking=LEVEL_1,
)



import logging

from parsl.data_provider.staging import Staging
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)


class NoOpFileStaging(Staging, RepresentationMixin):

    def can_stage_in(self, file):
        logger.debug("NoOpFileStaging checking file {}".format(repr(file)))
        return file.scheme == 'file'

    def can_stage_out(self, file):
        logger.debug("NoOpFileStaging checking file {}".format(repr(file)))
        return file.scheme == 'file'



from __future__ import annotations

import logging
import multiprocessing.queues as mpq
import os
import queue
import time
from multiprocessing.context import SpawnProcess as SpawnProcessType
from multiprocessing.queues import Queue as QueueType
from multiprocessing.synchronize import Event as EventType
from typing import Tuple

import typeguard
import zmq

from parsl.addresses import tcp_url
from parsl.log_utils import set_file_logger
from parsl.monitoring.errors import MonitoringRouterStartError
from parsl.monitoring.radios.multiprocessing import MultiprocessingQueueRadioSender
from parsl.monitoring.types import TaggedMonitoringMessage
from parsl.multiprocessing import (
    SpawnEvent,
    SpawnProcess,
    SpawnQueue,
    join_terminate_close_proc,
)
from parsl.process_loggers import wrap_with_logs
from parsl.utils import setproctitle

logger = logging.getLogger(__name__)


class MonitoringRouter:

    def __init__(self,
                 *,
                 address: str,
                 port_range: Tuple[int, int] = (55050, 56000),

                 run_dir: str = ".",
                 logging_level: int = logging.INFO,
                 resource_msgs: mpq.Queue,
                 exit_event: EventType,
                 ):
        """ Initializes a monitoring configuration class.

        Parameters
        ----------
        address : str
             The ip address at which the workers will be able to reach the Hub.
        port_range : tuple(int, int)
             The MonitoringHub picks ports at random from the range which will be used by Hub.
             Default: (55050, 56000)
        run_dir : str
             Parsl log directory paths. Logs and temp files go here. Default: '.'
        logging_level : int
             Logging level as defined in the logging module. Default: logging.INFO
        resource_msgs : multiprocessing.Queue
            A multiprocessing queue to receive messages to be routed onwards to the database process
        exit_event : Event
            An event that the main Parsl process will set to signal that the monitoring router should shut down.
        """
        os.makedirs(run_dir, exist_ok=True)
        set_file_logger(f"{run_dir}/monitoring_zmq_router.log",
                        level=logging_level)
        logger.debug("Monitoring router starting")

        self.address = address

        self.loop_freq = 10.0  # milliseconds

        self._context = zmq.Context()
        self.zmq_receiver_channel = self._context.socket(zmq.DEALER)
        self.zmq_receiver_channel.setsockopt(zmq.LINGER, 0)
        self.zmq_receiver_channel.set_hwm(0)
        self.zmq_receiver_channel.RCVTIMEO = int(self.loop_freq)  # in milliseconds
        logger.debug("address: {}. port_range {}".format(address, port_range))
        self.zmq_receiver_port = self.zmq_receiver_channel.bind_to_random_port(tcp_url(address),
                                                                               min_port=port_range[0],
                                                                               max_port=port_range[1])

        self.target_radio = MultiprocessingQueueRadioSender(resource_msgs)
        self.exit_event = exit_event

    @wrap_with_logs
    def start(self) -> None:
        logger.info("Starting ZMQ listener")
        try:
            while not self.exit_event.is_set():
                try:
                    dfk_loop_start = time.time()
                    while time.time() - dfk_loop_start < 1.0:  # TODO make configurable
                        # note that nothing checks that msg really is of the annotated type
                        msg: TaggedMonitoringMessage
                        msg = self.zmq_receiver_channel.recv_pyobj()

                        assert isinstance(msg, tuple), "ZMQ Receiver expects only tuples, got {}".format(msg)
                        assert len(msg) >= 1, "ZMQ Receiver expects tuples of length at least 1, got {}".format(msg)
                        assert len(msg) == 2, "ZMQ Receiver expects message tuples of exactly length 2, got {}".format(msg)

                        self.target_radio.send(msg)
                except zmq.Again:
                    pass
                except Exception:
                    # This will catch malformed messages. What happens if the
                    # channel is broken in such a way that it always raises
                    # an exception? Looping on this would maybe be the wrong
                    # thing to do.
                    logger.warning("Failure processing a ZMQ message", exc_info=True)

            logger.info("ZMQ listener finishing normally")
        finally:
            logger.info("ZMQ listener finished")


@wrap_with_logs
@typeguard.typechecked
def zmq_router_starter(*,
                       comm_q: mpq.Queue,
                       resource_msgs: mpq.Queue,
                       exit_event: EventType,

                       address: str,
                       port_range: Tuple[int, int],

                       run_dir: str,
                       logging_level: int) -> None:
    setproctitle("parsl: monitoring zmq router")
    try:
        router = MonitoringRouter(address=address,
                                  port_range=port_range,
                                  run_dir=run_dir,
                                  logging_level=logging_level,
                                  resource_msgs=resource_msgs,
                                  exit_event=exit_event)
    except Exception as e:
        logger.error("MonitoringRouter construction failed.", exc_info=True)
        comm_q.put(f"Monitoring router construction failed: {e}")
    else:
        comm_q.put(router.zmq_receiver_port)
        router.start()


class ZMQRadioReceiver():
    def __init__(self, *, process: SpawnProcessType, exit_event: EventType, port: int) -> None:
        self.process = process
        self.exit_event = exit_event
        self.port = port

    def close(self) -> None:
        self.exit_event.set()
        join_terminate_close_proc(self.process)


def start_zmq_receiver(*,
                       monitoring_messages: QueueType,
                       loopback_address: str,
                       port_range: Tuple[int, int],
                       logdir: str,
                       worker_debug: bool) -> ZMQRadioReceiver:
    comm_q = SpawnQueue(maxsize=10)

    router_exit_event = SpawnEvent()

    router_proc = SpawnProcess(target=zmq_router_starter,
                               kwargs={"comm_q": comm_q,
                                       "resource_msgs": monitoring_messages,
                                       "exit_event": router_exit_event,
                                       "address": loopback_address,
                                       "port_range": port_range,
                                       "run_dir": logdir,
                                       "logging_level": logging.DEBUG if worker_debug else logging.INFO,
                                       },
                               name="Monitoring-ZMQ-Router-Process",
                               daemon=True,
                               )
    router_proc.start()

    try:
        logger.debug("Waiting for router process to report port")
        comm_q_result = comm_q.get(block=True, timeout=120)
        comm_q.close()
        comm_q.join_thread()
    except queue.Empty:
        logger.error("Monitoring ZMQ Router has not reported port in 120s")
        raise MonitoringRouterStartError()

    if isinstance(comm_q_result, str):
        logger.error("MonitoringRouter sent an error message: %s", comm_q_result)
        raise RuntimeError(f"MonitoringRouter failed to start: {comm_q_result}")

    return ZMQRadioReceiver(process=router_proc, exit_event=router_exit_event, port=comm_q_result)



from parsl.app.app import join_app, python_app


@python_app
def add(*args):
    """Add all of the arguments together. If no arguments, then
    zero is returned (the neutral element of +)
    """
    accumulator = 0
    for v in args:
        accumulator += v
    return accumulator


@join_app
def fibonacci(n):
    if n == 0:
        return add()
    elif n == 1:
        return add(1)
    else:
        return add(fibonacci(n - 1), fibonacci(n - 2))


def test_fibonacci():
    assert fibonacci(0).result() == 0
    assert fibonacci(4).result() == 3
    assert fibonacci(6).result() == 8



from __future__ import annotations

import hashlib
import logging
import os
import pickle
import threading
import types
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future
from functools import lru_cache, singledispatch
from typing import Any, Dict, List, Literal, Optional, Sequence

import typeguard

from parsl.dataflow.errors import BadCheckpoint
from parsl.dataflow.taskrecord import TaskRecord
from parsl.errors import ConfigurationError
from parsl.utils import Timer, get_all_checkpoints

logger = logging.getLogger(__name__)


class CheckpointCommand:
    def __init__(self, task_record: TaskRecord, *, result: Optional[object] = None, exception: Optional[BaseException] = None):
        """Create a checkpoint command. This specifies a checkpoint entry of
        either a result or an exception. If exception is set, then this object
        specifies an exception checkpoint. Otherwise, it specifies a result.
        This is almost: one of exception or result must be non-None, but not
        quite because this data model also needs to accomodate a valid result
        of None."""

        assert result is None or exception is None, "CheckpointCommand cannot specify both a result and exception"

        self._task_record = task_record
        self._result = result
        self._exception = exception

    @property
    def task_record(self) -> TaskRecord:
        return self._task_record

    @property
    def result(self) -> Optional[object]:
        return self._result

    @property
    def exception(self) -> Optional[BaseException]:
        return self._exception


@singledispatch
def id_for_memo(obj: object, output_ref: bool = False) -> bytes:
    """This should return a byte sequence which identifies the supplied
    value for memoization purposes: for any two calls of id_for_memo,
    the byte sequence should be the same when the "same" value is supplied,
    and different otherwise.

    "same" is in quotes about because sameness is not as straightforward as
    serialising out the content.

    For example, for two dicts x, y:

      x = {"a":3, "b":4}
      y = {"b":4, "a":3}

    then: x == y, but their serialization is not equal, and some other
    functions on x and y are not equal: list(x.keys()) != list(y.keys())


    id_for_memo is invoked with output_ref=True when the parameter is an
    output reference (a value in the outputs=[] parameter of an app
    invocation).

    Memo hashing might be different for such parameters: for example, a
    user might choose to hash input File content so that changing the
    content of an input file invalidates memoization. This does not make
    sense to do for output files: there is no meaningful content stored
    where an output filename points at memoization time.
    """
    logger.error("id_for_memo attempted on unknown type {}".format(type(obj)))
    raise ValueError("unknown type for memoization: {}".format(type(obj)))


@id_for_memo.register(str)
@id_for_memo.register(int)
@id_for_memo.register(float)
@id_for_memo.register(type(None))
def id_for_memo_pickle(obj: object, output_ref: bool = False) -> bytes:
    return pickle.dumps(obj)


@id_for_memo.register(list)
def id_for_memo_list(denormalized_list: list, output_ref: bool = False) -> bytes:
    if type(denormalized_list) is not list:
        raise ValueError("id_for_memo_list cannot work on subclasses of list")

    normalized_list = []

    for e in denormalized_list:
        normalized_list.append(id_for_memo(e, output_ref=output_ref))

    return pickle.dumps(normalized_list)


@id_for_memo.register(tuple)
def id_for_memo_tuple(denormalized_tuple: tuple, output_ref: bool = False) -> bytes:
    if type(denormalized_tuple) is not tuple:
        raise ValueError("id_for_memo_tuple cannot work on subclasses of tuple")

    normalized_list = []

    for e in denormalized_tuple:
        normalized_list.append(id_for_memo(e, output_ref=output_ref))

    return pickle.dumps(normalized_list)


@id_for_memo.register(dict)
def id_for_memo_dict(denormalized_dict: dict, output_ref: bool = False) -> bytes:
    """This normalises the keys and values of the supplied dictionary.

    When output_ref=True, the values are normalised as output refs, but
    the keys are not.
    """
    if type(denormalized_dict) is not dict:
        raise ValueError("id_for_memo_dict cannot work on subclasses of dict")

    keys = sorted(denormalized_dict)

    normalized_list = []
    for k in keys:
        normalized_list.append(id_for_memo(k))
        normalized_list.append(id_for_memo(denormalized_dict[k], output_ref=output_ref))
    return pickle.dumps(normalized_list)


# the LRU cache decorator must be applied closer to the id_for_memo_function call
# that the .register() call, so that the cache-decorated version is registered.
@id_for_memo.register(types.FunctionType)
@lru_cache()
def id_for_memo_function(f: types.FunctionType, output_ref: bool = False) -> bytes:
    """This will checkpoint a function based only on its name and module name.
    This means that changing source code (other than the function name) will
    not cause a checkpoint invalidation.
    """
    return pickle.dumps(["types.FunctionType", f.__name__, f.__module__])


def make_hash(task: TaskRecord) -> str:
    """Create a hash of the task inputs.

    Args:
        - task (dict) : Task dictionary from dfk.tasks

    Returns:
        - hash (str) : A unique hash string
    """

    t: List[bytes] = []

    # if kwargs contains an outputs parameter, that parameter is removed
    # and normalised differently - with output_ref set to True.
    # kwargs listed in ignore_for_cache will also be removed

    filtered_kw = task['kwargs'].copy()

    ignore_list = task['ignore_for_cache']

    logger.debug("Ignoring these kwargs for checkpointing: %s", ignore_list)
    for k in ignore_list:
        logger.debug("Ignoring kwarg %s", k)
        del filtered_kw[k]

    if 'outputs' in task['kwargs']:
        outputs = task['kwargs']['outputs']
        del filtered_kw['outputs']
        t.append(id_for_memo(outputs, output_ref=True))

    t.extend(map(id_for_memo, (filtered_kw, task['func'], task['args'])))

    x = b''.join(t)
    return hashlib.md5(x).hexdigest()


class Memoizer(metaclass=ABCMeta):
    """Defines the interface for the DFK to talk to the memoization/checkpoint system.

    The DFK will invoke these methods on an instance of a Memoizer at suitable points
    in the lifecycle of a task. These methods are not intended for users to invoke
    directly.
    """

    @abstractmethod
    def update_memo_exception(self, task: TaskRecord, e: BaseException) -> None:
        """Called by the DFK when a task completes with an exception.

        On every task completion, either this method or `update_memo_result`
        will be called, but not both.

        This is an opportunity for the memoization/checkpoint system to record
        the outcome of a task for later discovery by a call to check_memo.
        """
        raise NotImplementedError

    @abstractmethod
    def update_memo_result(self, task: TaskRecord, r: Any) -> None:
        """Called by the DFK when a task completes with a successful result.

        On every task completion, either this method or `update_memo_exception`
        will be called, but not both.

        This is an opportunity for the memoization/checkpoint system to record
        the outcome of a task for later discovery by a call to check_memo.
        """
        raise NotImplementedError

    @abstractmethod
    def start(self, *, run_dir: str, config_run_dir: str) -> None:
        """Called by the DFK when it starts up.

        This is an opportunity for the memoization/checkpoint system to
        initialize itself.

        The path to the per-run run directory and the base run directory
        are passed as parameters.
        """
        raise NotImplementedError

    @abstractmethod
    def check_memo(self, task: TaskRecord) -> Optional[Future[Any]]:
        """Asks the checkpoint system for a result recorded for the described
        task. ``check_memo`` should return a `Future` that will be used as
        an executor future, in place of sending the task to an executor for
        execution. That future should be populated with a result or exception.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Called at DFK shutdown. This gives the checkpoint system an
        opportunity for graceful shutdown.
        """
        raise NotImplementedError


class BasicMemoizer(Memoizer):
    """Memoizer is responsible for ensuring that identical work is not repeated.

    When a task is repeated, i.e., the same function is called with the same exact arguments, the
    result from a previous execution is reused. `wiki <https://en.wikipedia.org/wiki/Memoization>`_

    The memoizer implementation here does not collapse duplicate calls
    at call time, but works **only** when the result of a previous
    call is available at the time the duplicate call is made.

    For instance::

       No advantage from                 Memoization helps
       memoization here:                 here:

        TaskA                            TaskB
          |   TaskA                        |
          |     |   TaskA                done  (TaskB)
          |     |     |                                (TaskB)
        done    |     |
              done    |
                    done

    The memoizer creates a lookup table by hashing the function name
    and its inputs, and storing the results of the function.

    When a task is ready for launch, i.e., all of its arguments
    have resolved, we add its hash to the task datastructure.

    """

    run_dir: str

    def __init__(self, *,
                 checkpoint_files: Sequence[str] | None = None,
                 checkpoint_period: Optional[str] = None,
                 checkpoint_mode: Literal['task_exit', 'periodic', 'dfk_exit', 'manual'] | None = None,
                 memoize: bool = True):
        """Initialize the memoizer.

        KWargs:

            - checkpoint_files : sequence of str, optional
                  List of paths to checkpoint files to load. By default, all checkpoints from the
                  run directory will be restored. This is usually the right behaviour, but this
                  parameter allows that behaviour to be overridden.
                  See :func:`parsl.utils.get_all_checkpoints` and :func:`parsl.utils.get_last_checkpoint`
                  for helpers.
            - checkpoint_period : str, optional
                  Time interval (in "HH:MM:SS") at which to checkpoint completed tasks. Only has an effect if
                  ``checkpoint_mode='periodic'``.
            - checkpoint_mode : str, optional
                  Checkpoint mode to use, can be ``'dfk_exit'``, ``'task_exit'``, ``'periodic'`` or ``'manual'``.
                  If set to `None`, checkpointing will be disabled. Default is None.
            - memoize : str, enable memoization or not.

        """
        self.checkpointed_tasks = 0

        # this lock must be held when:
        # * writing to any checkpoint files
        # * interacting with self.checkpointable_tasks
        self._checkpoint_lock = threading.Lock()

        self.checkpoint_files = checkpoint_files
        self.checkpoint_mode = checkpoint_mode
        self.checkpoint_period = checkpoint_period

        self.checkpointable_tasks: List[CheckpointCommand] = []

        self._checkpoint_timer: Timer | None = None
        self.memoize = memoize

    def start(self, *, run_dir: str, config_run_dir: str) -> None:

        self.run_dir = run_dir
        self.config_run_dir = config_run_dir

        if self.checkpoint_files is not None:
            checkpoint_files = self.checkpoint_files
        elif self.checkpoint_files is None and self.checkpoint_mode is not None:
            checkpoint_files = get_all_checkpoints(self.config_run_dir)
        else:
            checkpoint_files = []

        checkpoint = self.load_checkpoints(checkpoint_files)

        if self.memoize:
            logger.info("App caching initialized")
            self.memo_lookup_table = checkpoint
        else:
            logger.info("App caching disabled for all apps")
            self.memo_lookup_table = {}

        if self.checkpoint_mode == "periodic":
            if self.checkpoint_period is None:
                raise ConfigurationError("Checkpoint period must be specified with periodic checkpoint mode")
            else:
                try:
                    h, m, s = map(int, self.checkpoint_period.split(':'))
                except Exception:
                    raise ConfigurationError("invalid checkpoint_period provided: {0} expected HH:MM:SS".format(self.checkpoint_period))
                checkpoint_period = (h * 3600) + (m * 60) + s
                self._checkpoint_timer = Timer(self.checkpoint_queue, interval=checkpoint_period, name="Checkpoint")

    def close(self) -> None:
        if self.checkpoint_mode is not None:
            logger.info("Making final checkpoint")
            self.checkpoint_queue()

        if self._checkpoint_timer:
            logger.info("Stopping checkpoint timer")
            self._checkpoint_timer.close()

    def check_memo(self, task: TaskRecord) -> Optional[Future[Any]]:
        """Create a hash of the task and its inputs and check the lookup table for this hash.

        If present, the results are returned.

        Args:
            - task(task) : task from the dfk.tasks table

        Returns:
            - Result of the function if present in table, wrapped in a Future

        This call will also set task['hashsum'] to the unique hashsum for the func+inputs.
        """

        task_id = task['id']

        if not self.memoize or not task['memoize']:
            task['hashsum'] = None
            logger.debug("Task {} will not be memoized".format(task_id))
            return None

        hashsum = make_hash(task)
        logger.debug("Task {} has memoization hash {}".format(task_id, hashsum))
        result = None
        if hashsum in self.memo_lookup_table:
            result = self.memo_lookup_table[hashsum]
            logger.info("Task %s using result from cache", task_id)
        else:
            logger.info("Task %s had no result in cache", task_id)

        task['hashsum'] = hashsum

        assert isinstance(result, Future) or result is None
        return result

    def update_memo_result(self, task: TaskRecord, r: Any) -> None:
        self._update_memo(task)
        if self.checkpoint_mode is not None:
            self._update_checkpoint(CheckpointCommand(task, result=r))

    def update_memo_exception(self, task: TaskRecord, e: BaseException) -> None:
        self._update_memo(task)
        if self.checkpoint_mode is not None:
            self._update_checkpoint(CheckpointCommand(task, exception=e))

    def _update_memo(self, task: TaskRecord) -> None:
        """Updates the memoization lookup table with the result from a task.
        This doesn't move any values around but associates the memoization
        hashsum with the completed (by success or failure) AppFuture.

        Args:
             - task (TaskRecord) : A task record from dfk.tasks
        """
        task_id = task['id']

        if not self.memoize or not task['memoize'] or 'hashsum' not in task:
            return

        if not isinstance(task['hashsum'], str):
            logger.error("Attempting to update app cache entry but hashsum is not a string key")
            return

        if task['hashsum'] in self.memo_lookup_table:
            logger.info(f"Replacing app cache entry {task['hashsum']} with result from task {task_id}")
        else:
            logger.debug(f"Storing app cache entry {task['hashsum']} with result from task {task_id}")
        self.memo_lookup_table[task['hashsum']] = task['app_fu']

    def _load_checkpoints(self, checkpointDirs: Sequence[str]) -> Dict[str, Future[Any]]:
        """Load a checkpoint file into a lookup table.

        The data being loaded from the pickle file mostly contains input
        attributes of the task: func, args, kwargs, env...
        To simplify the check of whether the exact task has been completed
        in the checkpoint, we hash these input params and use it as the key
        for the memoized lookup table.

        Args:
            - checkpointDirs (list) : List of filepaths to checkpoints
              Eg. ['runinfo/001', 'runinfo/002']

        Returns:
            - memoized_lookup_table (dict)
        """
        memo_lookup_table = {}

        for checkpoint_dir in checkpointDirs:
            logger.info("Loading checkpoints from {}".format(checkpoint_dir))
            checkpoint_file = os.path.join(checkpoint_dir, 'tasks.pkl')
            try:
                with open(checkpoint_file, 'rb') as f:
                    while True:
                        try:
                            data = pickle.load(f)
                            # Copy and hash only the input attributes
                            memo_fu: Future = Future()
                            assert data['exception'] is None
                            memo_fu.set_result(data['result'])
                            memo_lookup_table[data['hash']] = memo_fu

                        except EOFError:
                            # Done with the checkpoint file
                            break
            except FileNotFoundError:
                reason = "Checkpoint file was not found: {}".format(
                    checkpoint_file)
                logger.error(reason)
                raise BadCheckpoint(reason)
            except Exception:
                reason = "Failed to load checkpoint: {}".format(
                    checkpoint_file)
                logger.error(reason)
                raise BadCheckpoint(reason)

            logger.info("Completed loading checkpoint: {0} with {1} tasks".format(checkpoint_file,
                                                                                  len(memo_lookup_table.keys())))
        return memo_lookup_table

    @typeguard.typechecked
    def load_checkpoints(self, checkpointDirs: Optional[Sequence[str]]) -> Dict[str, Future]:
        """Load checkpoints from the checkpoint files into a dictionary.

        The results are used to pre-populate the memoizer's lookup_table

        Kwargs:
             - checkpointDirs (list) : List of run folder to use as checkpoints
               Eg. ['runinfo/001', 'runinfo/002']

        Returns:
             - dict containing, hashed -> future mappings
        """
        if checkpointDirs:
            return self._load_checkpoints(checkpointDirs)
        else:
            return {}

    def _update_checkpoint(self, command: CheckpointCommand) -> None:
        if self.checkpoint_mode == 'task_exit':
            self.checkpoint_one(command)
        elif self.checkpoint_mode in ('manual', 'periodic', 'dfk_exit'):
            with self._checkpoint_lock:
                self.checkpointable_tasks.append(command)
        elif self.checkpoint_mode is None:
            pass
        else:
            assert False, "Invalid checkpoint mode {self.checkpoint_mode} - should have been validated at initialization"

    def checkpoint_one(self, cc: CheckpointCommand) -> None:
        """Checkpoint a single task to a checkpoint file.

        By default the checkpoints are written to the RUNDIR of the current
        run under RUNDIR/checkpoints/tasks.pkl

        Kwargs:
            - task : A task to checkpoint.

        .. note::
            Checkpointing only works if memoization is enabled

        """
        with self._checkpoint_lock:
            self._checkpoint_these_tasks([cc])

    def checkpoint_queue(self) -> None:
        """Checkpoint all tasks registered in self.checkpointable_tasks.

        By default the checkpoints are written to the RUNDIR of the current
        run under RUNDIR/checkpoints/tasks.pkl

        .. note::
            Checkpointing only works if memoization is enabled

        """
        with self._checkpoint_lock:
            self._checkpoint_these_tasks(self.checkpointable_tasks)
            self.checkpointable_tasks = []

    def checkpoint(self) -> None:
        """This is the user-facing interface to manual checkpointing.
        """
        self.checkpoint_queue()

    def _checkpoint_these_tasks(self, checkpoint_queue: List[CheckpointCommand]) -> None:
        """Play a sequence of CheckpointCommands into a checkpoint file.

        The checkpoint lock must be held when invoking this method.
        """

        # This checks that the lock is held, at least, but does not check that
        # it is held by the current thread - threading.Lock does not have a
        # concept of locking thread for threading.Lock.
        assert self._checkpoint_lock.locked(), "checkpoint system should be locked"

        checkpoint_dir = '{0}/checkpoint'.format(self.run_dir)
        checkpoint_tasks = checkpoint_dir + '/tasks.pkl'

        if not os.path.exists(checkpoint_dir):
            os.makedirs(checkpoint_dir, exist_ok=True)

        count = 0

        with open(checkpoint_tasks, 'ab') as f:
            for cc in checkpoint_queue:
                if cc.exception is None:
                    hashsum = cc.task_record['hashsum']
                    if not hashsum:
                        continue
                    t = {'hash': hashsum, 'exception': None, 'result': cc.result}

                    # We are using pickle here since pickle dumps to a file in 'ab'
                    # mode behave like a incremental log.
                    pickle.dump(t, f)
                    count += 1
                    logger.debug("Task %s checkpointed", cc.task_record['id'])

        self.checkpointed_tasks += count

        if count == 0:
            if self.checkpointed_tasks == 0:
                logger.warning("No tasks checkpointed so far in this run. Please ensure caching is enabled")
            else:
                logger.debug("No tasks checkpointed in this pass.")
        else:
            logger.info("Done checkpointing {} tasks".format(count))



import logging
import pickle
from contextlib import ExitStack
from typing import Optional

import zmq

from parsl import curvezmq

logger = logging.getLogger(__name__)


def probe_addresses(
    zmq_context: curvezmq.ClientContext,
    addresses: str,
    timeout_ms: int = 120_000,
    identity: Optional[bytes] = None,
):
    """
    Given a single-line CSV list of addresses, return the first proven valid address.

    This function will connect to each address in ``addresses`` (a comma-separated
    list of URLs) and attempt to send a CONNECTION_PROBE packet.  Returns the first
    address that receives a response.

    If no address receives a response within the ``timeout_ms`` (specified in
    milliseconds), then raise ``ConnectionError``.

    :param zmq_context: A ZMQ Context; the call-site may provide an encrypted ZMQ
        context for assurance that the returned address is the expected and correct
        endpoint
    :param addresses: a comma-separated string of addresses to attempt.  Example:
        ``tcp://127.0.0.1:1234,tcp://[3812::03aa]:5678``
    :param timeout_ms: how long to wait for a response from the probes.  The probes
        are initiated and await concurrently, so this timeout will be the total wall
        time in the worst case of "no addresses are valid."
    :param identity: a ZMQ connection identity; used for logging connection probes
        at the interchange
    :raises: ``ConnectionError`` if no addresses are determined valid
    :returns: a single address, the first one that received a response
    """
    if not addresses:
        raise ValueError("No address to probe!")

    sock_map = {}
    urls = addresses.split(",")
    with ExitStack() as stk:
        poller = zmq.Poller()
        for url in urls:
            logger.debug("Testing ZMQ connection to url: %s", url)
            s: zmq.Socket = stk.enter_context(zmq_context.socket(zmq.DEALER))
            s.setsockopt(zmq.LINGER, 0)
            s.setsockopt(zmq.IPV6, True)
            if identity:
                s.setsockopt(zmq.IDENTITY, identity)
            stk.enter_context(s.connect(url))
            poller.register(s, zmq.POLLIN)
            sock_map[s] = url

            s.send(pickle.dumps({'type': 'connection_probe'}))

        for sock, evt in poller.poll(timeout=timeout_ms):
            sock.recv()  # clear the buffer for good netizenry
            return sock_map.get(sock)

    addys = ", ".join(urls)  # just slightly more human friendly
    raise ConnectionError(f"No viable ZMQ url from: {addys}")


class TestWorker:

    def __init__(self, addresses, port):
        uid = str(uuid.uuid4())
        self.context = zmq.Context()
        self.task_incoming = self.context.socket(zmq.DEALER)
        self.task_incoming.setsockopt(zmq.IDENTITY, uid.encode('utf-8'))
        # Linger is set to 0, so that the manager can exit even when there might be
        # messages in the pipe
        self.task_incoming.setsockopt(zmq.LINGER, 0)

        address = probe_addresses(addresses, port)
        print("Viable address :", address)
        self.task_incoming.connect(tcp_url(address, port))

    def heartbeat(self):
        """ Send heartbeat to the incoming task queue
        """
        HEARTBEAT_CODE = (2 ** 32) - 1
        heartbeat = (HEARTBEAT_CODE).to_bytes(4, "little")
        r = self.task_incoming.send(heartbeat)
        print("Return from heartbeat: {}".format(r))


if __name__ == "__main__":
    import argparse
    import uuid

    from parsl.addresses import get_all_addresses, tcp_url

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", required=True,
                        help="Port to connect to")

    args = parser.parse_args()
    addresses = get_all_addresses()
    worker = TestWorker(addresses, args.port)
    worker.heartbeat()



# Cloud Providers
from parsl.providers.aws.aws import AWSProvider
from parsl.providers.azure.azure import AzureProvider
from parsl.providers.condor.condor import CondorProvider
from parsl.providers.googlecloud.googlecloud import GoogleCloudProvider
from parsl.providers.grid_engine.grid_engine import GridEngineProvider

# Kubernetes
from parsl.providers.kubernetes.kube import KubernetesProvider
from parsl.providers.local.local import LocalProvider
from parsl.providers.lsf.lsf import LSFProvider
from parsl.providers.pbspro.pbspro import PBSProProvider
from parsl.providers.slurm.slurm import SlurmProvider
from parsl.providers.torque.torque import TorqueProvider

__all__ = ['LocalProvider',
           'CondorProvider',
           'GridEngineProvider',
           'SlurmProvider',
           'TorqueProvider',
           'LSFProvider',
           'PBSProProvider',
           'AWSProvider',
           'GoogleCloudProvider',
           'KubernetesProvider',
           'AzureProvider']



from parsl.app.app import python_app


@python_app(cache=True, ignore_for_cache=[])
def random_uuid(x):
    import uuid
    return str(uuid.uuid4())


@python_app(cache=True, ignore_for_cache=['x'])
def random_uuid_def(x):
    import uuid
    return str(uuid.uuid4())


def test_memo_different():
    # explicitly call result() before launching next app so that
    # the executions are serialized and the x result will be in
    # memo table
    x = random_uuid(x=0).result()
    y = random_uuid(x=1).result()

    assert x != y, "Memoized results were used incorrectly"


def test_memo_same_at_definition():
    x = random_uuid_def(x=0).result()
    y = random_uuid_def(x=1).result()

    assert x == y, "Memoized results were not used"



import random
import string

import pytest

from parsl.providers.errors import SubmitException


@pytest.mark.local
def test_submit_exception_task_name_deprecation():
    """This tests the deprecation warning of task_name in SubmitException
    """
    j = "the_name-" + "".join(random.sample(string.ascii_lowercase, 10))

    ex = SubmitException(j, "m")

    # the new behaviour
    assert ex.job_name == j

    # the old behaviour
    with pytest.deprecated_call():
        assert ex.task_name == j



import random
import zipfile

import pytest

import parsl
from parsl.config import Config
from parsl.data_provider.files import File
from parsl.data_provider.zip import ZipAuthorityError, ZipFileStaging
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import LocalProvider
from parsl.tests.configs.htex_local import fresh_config as local_config


@parsl.python_app
def count_lines(file):
    with open(file, "r") as f:
        return len(f.readlines())


@pytest.mark.local
def test_zip_in(tmpd_cwd):
    # basic test of zip file stage-in
    zip_path = tmpd_cwd / "container.zip"
    file_base = "data.txt"
    zip_file = File(f"zip:{zip_path / file_base}")

    # create a zip file containing one file with some abitrary number of lines
    n_lines = random.randint(0, 1000)

    with zipfile.ZipFile(zip_path, mode='w') as z:
        with z.open(file_base, mode='w') as f:
            for _ in range(n_lines):
                f.write(b'someline\n')

    app_future = count_lines(zip_file)

    assert app_future.result() == n_lines



import logging
import pathlib
from typing import Optional
from unittest import mock

import psutil
import pytest
import zmq

from parsl import curvezmq
from parsl.executors.high_throughput.interchange import Interchange
from parsl.executors.high_throughput.manager_selector import RandomManagerSelector


def make_interchange(*,
                     interchange_address: Optional[str],
                     cert_dir: Optional[str],
                     worker_port: Optional[int] = None) -> Interchange:
    return Interchange(interchange_address=interchange_address,
                       cert_dir=cert_dir,
                       client_address="127.0.0.1",
                       client_ports=(50055, 50056, 50057),
                       worker_port=worker_port,
                       worker_port_range=(54000, 55000),
                       hub_address=None,
                       hub_zmq_port=None,
                       heartbeat_threshold=60,
                       logdir=".",
                       logging_level=logging.INFO,
                       manager_selector=RandomManagerSelector(),
                       poll_period=10,
                       run_id="test_run_id",
                       _check_python_mismatch=True)


@pytest.fixture
def encrypted(request: pytest.FixtureRequest):
    if hasattr(request, "param"):
        return request.param
    return True


@pytest.fixture
def cert_dir(encrypted: bool, tmpd_cwd: pathlib.Path):
    if not encrypted:
        return None
    return curvezmq.create_certificates(tmpd_cwd)


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
@mock.patch.object(curvezmq.ServerContext, "socket", return_value=mock.MagicMock())
def test_interchange_curvezmq_sockets(
    mock_socket: mock.MagicMock, cert_dir: Optional[str], encrypted: bool
):
    address = "127.0.0.1"
    ix = make_interchange(interchange_address=address, cert_dir=cert_dir)
    assert isinstance(ix.zmq_context, curvezmq.ServerContext)
    assert ix.zmq_context.encrypted is encrypted
    assert mock_socket.call_count == 4


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_interchange_binding_no_address(cert_dir: Optional[str]):
    ix = make_interchange(interchange_address=None, cert_dir=cert_dir)
    assert ix.interchange_address == "*"


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_interchange_binding_with_address(cert_dir: Optional[str]):
    # Using loopback address
    address = "127.0.0.1"
    ix = make_interchange(interchange_address=address, cert_dir=cert_dir)
    assert ix.interchange_address == address


@pytest.mark.skip("This behaviour is possibly unexpected. See issue #3037")
@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_interchange_binding_with_non_ipv4_address(cert_dir: Optional[str]):
    # Confirm that a ipv4 address is required
    address = "localhost"
    with pytest.raises(zmq.error.ZMQError):
        make_interchange(interchange_address=address, cert_dir=cert_dir)


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_interchange_binding_bad_address(cert_dir: Optional[str]):
    """Confirm that we raise a ZMQError when a bad address is supplied"""
    address = "550.0.0.0"
    with pytest.raises(ValueError):
        make_interchange(interchange_address=address, cert_dir=cert_dir)


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_limited_interface_binding(cert_dir: Optional[str]):
    """When address is specified the worker_port would be bound to it rather than to 0.0.0.0"""
    address = "127.0.0.1"
    ix = make_interchange(interchange_address=address, cert_dir=cert_dir)
    proc = psutil.Process()
    conns = proc.connections(kind="tcp")

    matched_conns = [conn for conn in conns if conn.laddr.port == ix.worker_port]
    assert len(matched_conns) == 1
    # laddr.ip can return ::ffff:127.0.0.1 when using IPv6
    assert address in matched_conns[0].laddr.ip


@pytest.mark.local
@pytest.mark.parametrize("encrypted", (True, False), indirect=True)
def test_fixed_ports(cert_dir: Optional[str]):
    ix = make_interchange(interchange_address=None, cert_dir=cert_dir, worker_port=51117)
    assert ix.interchange_address == "*"



import zmq

from parsl.monitoring.radios.base import MonitoringRadioSender


class ZMQRadioSender(MonitoringRadioSender):
    """A monitoring radio which connects over ZMQ. This radio is not
    thread-safe, because its use of ZMQ is not thread-safe.
    """

    def __init__(self, hub_address: str, hub_zmq_port: int) -> None:
        self._hub_channel = zmq.Context().socket(zmq.DEALER)
        self._hub_channel.set_hwm(0)
        self._hub_channel.connect(f"tcp://{hub_address}:{hub_zmq_port}")

    def send(self, message: object) -> None:
        self._hub_channel.send_pyobj(message)



from parsl.addresses import address_by_route
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import KubernetesProvider
from parsl.usage_tracking.levels import LEVEL_1

config = Config(
    executors=[
        HighThroughputExecutor(
            label='kube-htex',
            cores_per_worker=1,
            max_workers_per_node=1,
            worker_logdir_root='YOUR_WORK_DIR',

            # Address for the pod worker to connect back
            address=address_by_route(),
            provider=KubernetesProvider(
                namespace="default",

                # Docker image url to use for pods
                image='YOUR_DOCKER_URL',

                # Command to be run upon pod start, such as:
                # 'module load Anaconda; source activate parsl_env'.
                # or 'pip install parsl'
                worker_init='',

                # The secret key to download the image
                secret="YOUR_KUBE_SECRET",

                # Should follow the Kubernetes naming rules
                pod_name='YOUR-POD-Name',

                nodes_per_block=1,
                init_blocks=1,
                # Maximum number of pods to scale up
                max_blocks=10,
            ),
        ),
    ],
    usage_tracking=LEVEL_1,
)



from parsl.config import Config
from parsl.data_provider.file_noop import NoOpFileStaging
from parsl.data_provider.ftp import FTPInTaskStaging
from parsl.data_provider.http import HTTPInTaskStaging
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import LocalProvider

config = Config(
    executors=[
        HighThroughputExecutor(
            label="htex_Local",
            worker_debug=True,
            cores_per_worker=1,
            encrypted=True,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
                launcher=SimpleLauncher(),
            ),
            storage_access=[HTTPInTaskStaging(), FTPInTaskStaging(), NoOpFileStaging()]
        )
    ],
    strategy='none',
)



import logging
import os
import time

from parsl.jobs.states import JobState, JobStatus
from parsl.launchers import AprunLauncher
from parsl.providers.cluster_provider import ClusterProvider
from parsl.providers.torque.template import template_string
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)

# From the man pages for qstat for PBS/Torque systems
translate_table = {
    'B': JobState.RUNNING,  # This state is returned for running array jobs
    'R': JobState.RUNNING,
    'C': JobState.COMPLETED,  # Completed after having run
    'E': JobState.COMPLETED,  # Exiting after having run
    'H': JobState.HELD,  # Held
    'Q': JobState.PENDING,  # Queued, and eligible to run
    'W': JobState.PENDING,  # Job is waiting for it's execution time (-a option) to be reached
    'S': JobState.HELD  # Suspended
}


class TorqueProvider(ClusterProvider, RepresentationMixin):
    """Torque Execution Provider

    This provider uses qsub to submit, qstat for status, and qdel to cancel
    jobs. The qsub script to be used is created from a template file in this
    same module.

    Parameters
    ----------
    account : str
        Account the job will be charged against.
    queue : str
        Torque queue to request blocks from.
    nodes_per_block : int
        Nodes to provision per block.
    init_blocks : int
        Number of blocks to provision at the start of the run. Default is 1.
    min_blocks : int
        Minimum number of blocks to maintain. Default is 0.
    max_blocks : int
        Maximum number of blocks to maintain.
    parallelism : float
        Ratio of provisioned task slots to active tasks. A parallelism value of 1 represents aggressive
        scaling where as many resources as possible are used; parallelism close to 0 represents
        the opposite situation in which as few resources as possible (i.e., min_blocks) are used.
    walltime : str
        Walltime requested per block in HH:MM:SS.
    scheduler_options : str
        String to prepend to the #PBS blocks in the submit script to the scheduler.
        WARNING: scheduler_options should only be given #PBS strings, and should not have trailing newlines.
    worker_init : str
        Command to be run before starting a worker, such as 'module load Anaconda; source activate env'.
    launcher : Launcher
        Launcher for this provider. Possible launchers include
        :class:`~parsl.launchers.AprunLauncher` (the default), or
        :class:`~parsl.launchers.SingleNodeLauncher`,

    """
    def __init__(self,
                 account=None,
                 queue=None,
                 scheduler_options='',
                 worker_init='',
                 nodes_per_block=1,
                 init_blocks=1,
                 min_blocks=0,
                 max_blocks=1,
                 parallelism=1,
                 launcher=AprunLauncher(),
                 walltime="00:20:00",
                 cmd_timeout=120):
        label = 'torque'
        super().__init__(label,
                         nodes_per_block,
                         init_blocks,
                         min_blocks,
                         max_blocks,
                         parallelism,
                         walltime,
                         launcher,
                         cmd_timeout=cmd_timeout)

        self.account = account
        self.queue = queue
        self.scheduler_options = scheduler_options
        self.worker_init = worker_init
        self.template_string = template_string

        # Dictionary that keeps track of jobs, keyed on job_id
        self.resources = {}

    def _status(self):
        '''Returns the status list for a list of job_ids

        Args:
              self

        Returns:
              [status...] : Status list of all jobs
        '''

        job_ids = list(self.resources.keys())
        job_id_list = ' '.join(self.resources.keys())

        jobs_missing = list(self.resources.keys())

        retcode, stdout, stderr = self.execute_wait("qstat {0}".format(job_id_list))
        for line in stdout.split('\n'):
            parts = line.split()
            if not parts or parts[0].upper().startswith('JOB') or parts[0].startswith('---'):
                continue
            job_id = parts[0]  # likely truncated
            for long_job_id in job_ids:
                if long_job_id.startswith(job_id):
                    logger.debug('coerced job_id %s -> %s', job_id, long_job_id)
                    job_id = long_job_id
                    break
            state = translate_table.get(parts[4], JobState.UNKNOWN)
            self.resources[job_id]['status'] = JobStatus(state)
            jobs_missing.remove(job_id)

        # squeue does not report on jobs that are not running. So we are filling in the
        # blanks for missing jobs, we might lose some information about why the jobs failed.
        for missing_job in jobs_missing:
            self.resources[missing_job]['status'] = JobStatus(JobState.COMPLETED)

    def submit(self, command, tasks_per_node, job_name="parsl.torque"):
        ''' Submits the command onto an Local Resource Manager job.
        Submit returns an ID that corresponds to the task that was just submitted.

        If tasks_per_node <  1 : ! This is illegal. tasks_per_node should be integer

        If tasks_per_node == 1:
             A single node is provisioned

        If tasks_per_node >  1 :
             tasks_per_node number of nodes are provisioned.

        Args:
             - command  :(String) Commandline invocation to be made on the remote side.
             - tasks_per_node (int) : command invocations to be launched per node

        Kwargs:
             - job_name (String): Name for job, must be unique

        Returns:
             - None: At capacity, cannot provision more
             - job_id: (string) Identifier for the job

        '''

        # Set job name
        job_name = "parsl.{0}.{1}".format(job_name, time.time())

        # Set script path
        script_path = "{0}/{1}.submit".format(self.script_dir, job_name)
        script_path = os.path.abspath(script_path)

        logger.debug("Requesting nodes_per_block:%s tasks_per_node:%s", self.nodes_per_block,
                     tasks_per_node)

        job_config = {}
        job_config["submit_script_dir"] = self.script_dir
        job_config["nodes"] = self.nodes_per_block
        job_config["task_blocks"] = self.nodes_per_block * tasks_per_node
        job_config["nodes_per_block"] = self.nodes_per_block
        job_config["tasks_per_node"] = tasks_per_node
        job_config["walltime"] = self.walltime
        job_config["scheduler_options"] = self.scheduler_options
        job_config["worker_init"] = self.worker_init
        job_config["user_script"] = command

        # Wrap the command
        job_config["user_script"] = self.launcher(command,
                                                  tasks_per_node,
                                                  self.nodes_per_block,
                                                  self.script_dir)

        logger.debug("Writing submit script")
        self._write_submit_script(self.template_string, script_path, job_name, job_config)

        submit_options = ''
        if self.queue is not None:
            submit_options = '{0} -q {1}'.format(submit_options, self.queue)
        if self.account is not None:
            submit_options = '{0} -A {1}'.format(submit_options, self.account)

        launch_cmd = "qsub {0} {1}".format(submit_options, script_path)
        retcode, stdout, stderr = self.execute_wait(launch_cmd)

        job_id = None
        if retcode == 0:
            for line in stdout.split('\n'):
                if line.strip():
                    job_id = line.strip()
                    self.resources[job_id] = {'job_id': job_id, 'status': JobStatus(JobState.PENDING)}
        else:
            message = "Command '{}' failed with return code {}".format(launch_cmd, retcode)
            if (stdout is not None) and (stderr is not None):
                message += "\nstderr:{}\nstdout{}".format(stderr.strip(), stdout.strip())
            logger.error(message)

        return job_id

    def cancel(self, job_ids):
        ''' Cancels the jobs specified by a list of job ids

        Args:
        job_ids : [<job_id> ...]

        Returns :
        [True/False...] : If the cancel operation fails the entire list will be False.
        '''

        job_id_list = ' '.join(job_ids)
        retcode, stdout, stderr = self.execute_wait("qdel {0}".format(job_id_list))
        rets = None
        if retcode == 0:
            for jid in job_ids:
                self.resources[jid]['status'] = JobStatus(JobState.COMPLETED)  # Setting state to exiting
            rets = [True for i in job_ids]
        else:
            rets = [False for i in job_ids]

        return rets

    @property
    def status_polling_interval(self):
        return 60



from parsl.executors.flux.executor import FluxExecutor
from parsl.executors.globus_compute import GlobusComputeExecutor
from parsl.executors.high_throughput.executor import HighThroughputExecutor
from parsl.executors.high_throughput.mpi_executor import MPIExecutor
from parsl.executors.threads import ThreadPoolExecutor
from parsl.executors.workqueue.executor import WorkQueueExecutor

__all__ = ['ThreadPoolExecutor',
           'HighThroughputExecutor',
           'MPIExecutor',
           'WorkQueueExecutor',
           'FluxExecutor',
           'GlobusComputeExecutor']



"""Tests monitoring records app name under various decoration patterns.
"""

import os
import time

import pytest

import parsl
from parsl.tests.configs.htex_local_alternate import fresh_config


@parsl.python_app
def regular_decorated_app():
    return 5


@pytest.mark.local
def get_regular_decorated_app():
    return regular_decorated_app


def for_decoration_later():
    return 77


def get_for_decoration_later():
    return parsl.python_app(for_decoration_later)


def get_decorated_closure():

    r = 53

    @parsl.python_app
    def decorated_closure():
        return r

    return decorated_closure


@pytest.mark.local
@pytest.mark.parametrize("get_app,expected_name,expected_result",
                         [(get_regular_decorated_app, "regular_decorated_app", 5),
                          (get_for_decoration_later, "for_decoration_later", 77),
                          (get_decorated_closure, "decorated_closure", 53)
                          ])
def test_app_name(get_app, expected_name, expected_result, tmpd_cwd):

    # this is imported here rather than at module level because
    # it isn't available in a plain parsl install, so this module
    # would otherwise fail to import and break even a basic test
    # run.
    import sqlalchemy

    c = fresh_config()
    c.run_dir = str(tmpd_cwd)
    c.monitoring.logging_endpoint = f"sqlite:///{tmpd_cwd}/monitoring.db"
    parsl.load(c)

    app = get_app()
    assert app().result() == expected_result

    parsl.dfk().cleanup()

    engine = sqlalchemy.create_engine(c.monitoring.logging_endpoint)
    with engine.begin() as connection:

        def count_rows(table: str):
            result = connection.execute(sqlalchemy.text(f"SELECT COUNT(*) FROM {table}"))
            (c, ) = result.first()
            return c

        # one workflow...
        assert count_rows("workflow") == 1

        # ... with one task ...
        assert count_rows("task") == 1

        # ... that was tried once ...
        assert count_rows("try") == 1

        # ... and has the expected name.
        result = connection.execute(sqlalchemy.text("SELECT task_func_name FROM task"))
        (c, ) = result.first()
        assert c == expected_name



"""Script meant to be the initial program of a Flux instance."""

import argparse
import logging
import os
from os.path import dirname
from socket import gethostbyname, gethostname

import zmq


def main():
    """Run a Flux instance to completion.

    Send the path to the Flux Python package and the URI of the
    encapsulating Flux instance.
    """
    # flux imports only available when launched under Flux instance
    import flux
    import flux.job

    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s [%(levelname)s] %(message)s"
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("protocol", help="Protocol of the parent executor's socket")
    parser.add_argument("hostname", help="hostname of the parent executor's socket")
    parser.add_argument("port", help="Port of the parent executor's socket")
    args = parser.parse_args()
    with zmq.Context() as context, context.socket(zmq.REQ) as socket:
        socket.connect(
            args.protocol + "://" + gethostbyname(args.hostname) + ":" + args.port
        )
        # send the path to the ``flux.job`` package
        socket.send(dirname(dirname(os.path.realpath(flux.__file__))).encode())
        logging.debug("Flux package path sent.")
        # collect the encapsulating Flux instance's URI
        local_uri = flux.Flux().attr_get("local-uri")
        hostname = gethostname()
        if args.hostname == hostname:
            flux_uri = local_uri
        else:
            flux_uri = "ssh://" + gethostname() + local_uri.replace("local://", "")
        logging.debug("Flux URI is %s", flux_uri)
        response = socket.recv()  # get acknowledgment
        logging.debug("Received acknowledgment %s", response)
        socket.send(flux_uri.encode())  # send URI
        logging.debug("URI sent. Blocking for response...")
        response = socket.recv()  # wait for shutdown message
        logging.debug("Response %s received, draining flux jobs...", response)
        flux.Flux().rpc("job-manager.drain").get()
        logging.debug("Flux jobs drained, exiting.")


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional

# This try except clause prevents import errors
# when TaskVine is not used in Parsl.
try:
    from ndcctools.taskvine.cvine import VINE_DEFAULT_PORT
except ImportError:
    VINE_DEFAULT_PORT = 0   # use any available port.


@dataclass
class TaskVineManagerConfig:
    """
    Configuration of a TaskVine manager

    Parameters
    ----------

    port: int
        Network port for the manager to communicate with workers.
        A value of 0 means TaskVine chooses any available port.
        Default is VINE_DEFAULT_PORT.

    address: Optional[str]
        Address of the local machine.
        If None, :py:func:`parsl.addresses.get_any_address` will be used to determine the address.

    project_name: Optional[str]
        If given, TaskVine will periodically report its status and performance
        back to the global TaskVine catalog at
        http://ccl.cse.nd.edu/software/taskvine/status.
        Recommended mode of communication between manager and workers.
        Default is None. Overrides address.

    project_password_file: Optional[str]
        If given, manager and workers will authenticate each other
        with the given password file.
        Default is None.

    env_vars: Optional[dict]
        Dictionary of environment variables to set in a shell
        before executing a task.
        Default is None

    init_command: str
        Command line to run before executing a task in a worker.
        Default is ''.

    env_pack: Optional[str]
        Used to encapsulate package dependencies of tasks to
        execute them remotely without needing a shared filesystem.
        Recommended way to manage tasks' dependency requirements.
        All tasks will be executed in the encapsulated environment.
        If an absolute path to a conda environment or a conda
        environment name is given, TaskVine will package the conda
        environment in a tarball and send it along with tasks to be
        executed in a replicated conda environment.
        If a tarball of packages (``*.tar.gz``) is given, TaskVine
        skips the packaging step and sends the tarball along with
        tasks to be executed in a replicated conda environment.

    app_pack: bool
        Use conda-pack to prepare a self-contained Python environment
        for each app. Enabling this increases first task latency but
        does not require a common environment or a shared filesystem
        on workers.
        If env_pack is specified, app_pack if set to True will override
        env_pack.
        Default is False.

    extra_pkgs: Optional[list]
        Used when app_pack is True.
        List of extra pip/conda package names to include when packing
        the environment. Useful when app executes other (possibly
        non-Python) programs provided via pip or conda.
        Default is None.

    max_retries: int
        Set the number of retries that TaskVine will make when a task
        fails. This is distinct from Parsl level retries configured in
        parsl.config.Config. Set to None to allow TaskVine to retry tasks
        forever.
        Default is 1.

    library_config: Optional[dict]
        Only and must specify when functions are executed in the serverless mode.
        Configure the number of function slots and amount of resources
        a library task can run. A library task is a stateful object that executes
        functions in the serverless way. Accept the following keywords:
        'num_slots', 'cores', 'memory (MBs)', 'disk (MBs)'.
        Default is {'num_slots': 1, 'cores': None, 'memory': None, 'disk': None},
        which will take all resources of a worker node and run at most 1 function
        invocation at any given time.
        E.g., {'num_slots': 4, 'cores': 16, 'memory': 16000, 'disk': 16000} will
        reserve those resources to the library task to run at most 4 function
        invocations.

    shared_fs: bool
        Whether workers will use a shared filesystem or not. If so, TaskVine
        will not track and transfer files for execution, in exchange for
        I/O pressure on the shared filesystem.
        Default is False.

    autolabel: bool
        Use the Resource Monitor to automatically determine resource
        labels based on observed task behavior.
        Default is False.

    autolabel_window: int
        Set the number of tasks considered for autolabeling. TaskVine
        will wait for a series of N tasks with steady resource
        requirements before making a decision on labels. Increasing
        this parameter will reduce the number of failed tasks due to
        resource exhaustion when autolabeling, at the cost of increased
        resources spent collecting stats.

    autolabel_algorithm: str
        Choose an autolabeling algorithm to label resources to tasks.
        'max-xput': smartly guess a resource value, then fall back if task
        fails due to overconsumption.
        'bucketing': group tasks into buckets based on resource consumption
        similarity, and randomly choose a bucket to assign resources to the
        next task.
        'max': always allocate max resources seen so far. Allocate a whole
        machine if task fails.
        Default is max-xput.

    autolabel_window: Optional[int]
        Only apply to 'max-xput' and 'max'.
        Set the number of tasks considered for autolabeling. TaskVine
        will wait for a series of N tasks with steady resource
        requirements before making a decision on labels. Increasing
        this parameter will reduce the number of failed tasks due to
        resource exhaustion when autolabeling, at the cost of increased
        resources spent collecting stats.

    autocategory: bool
        Place each app in its own category by default. If all
        invocations of an app have similar performance characteristics,
        this will provide a reasonable set of categories automatically.
        Default is True.

    enable_peer_transfers: bool
        Whether to enable workers' capability to transfer files among
        themselves or not. Helpful if TaskVine manages all files
        tasks in the workflow use. Require that workers can communicate
        with each other via TCP ports.
        Default is True.

    wait_for_workers: Optional[int]
        Number of workers to wait for before the manager starts sending
        tasks for execution.

    vine_log_dir: Optional[str]
        Directory to store TaskVine logging facilities.
        Default is None, in which all TaskVine logs will be contained
        in the Parsl logging directory.

    tune_parameters: Optional[dict]
        Extended vine_tune parameters, expressed in a dictionary
        by { 'tune-parameter' : value }.
    """

    # Connection and communication settings
    port: int = VINE_DEFAULT_PORT
    address: Optional[str] = None
    project_name: Optional[str] = None
    project_password_file: Optional[str] = None

    # Global task settings
    env_vars: Optional[dict] = None
    init_command: str = ""
    env_pack: Optional[str] = None
    app_pack: bool = False
    extra_pkgs: Optional[list] = None
    max_retries: int = 1
    library_config: Optional[dict] = None

    # Performance-specific settings
    shared_fs: bool = False
    autolabel: bool = False
    autolabel_algorithm: str = 'max-xput'
    autolabel_window: Optional[int] = None
    autocategory: bool = True
    enable_peer_transfers: bool = True
    wait_for_workers: Optional[int] = None
    tune_parameters: Optional[dict] = None

    # Logging settings
    vine_log_dir: Optional[str] = None



import pytest

import parsl


@parsl.python_app
def always_fails():
    raise ValueError("always_fails deliberate exception")


def retry_handler_raises(exc, task_record):
    raise RuntimeError("retry_handler_raises deliberate exception")


def local_config():
    return parsl.config.Config(retry_handler=retry_handler_raises)


@pytest.mark.local
def test_retry_handler_exception():
    fut = always_fails()
    with pytest.raises(RuntimeError):
        fut.result()
    assert fut.exception().args[0] == "retry_handler_raises deliberate exception"



import zipfile

import pytest

import parsl
from parsl.config import Config
from parsl.data_provider.data_manager import default_staging
from parsl.data_provider.files import File
from parsl.data_provider.zip import ZipAuthorityError, ZipFileStaging
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import LocalProvider
from parsl.tests.configs.htex_local import fresh_config as local_config


@pytest.mark.local
def test_zip_path_split():
    from parsl.data_provider.zip import zip_path_split
    assert zip_path_split("/tmp/foo/this.zip/inside/here.txt") == ("/tmp/foo/this.zip", "inside/here.txt")


@parsl.bash_app
def output_something(outputs=()):
    """This should output something into every specified output file:
    the position in the output sequence will be written into the
    corresponding output file.
    """
    cmds = []
    for n in range(len(outputs)):
        cmds.append(f"echo {n} > {outputs[n]}")

    return "; ".join(cmds)


@pytest.mark.local
def test_zip_out(tmpd_cwd):
    # basic test of zip file stage-out
    zip_path = tmpd_cwd / "container.zip"
    file_base = "data.txt"
    of = File(f"zip:{zip_path / file_base}")

    app_future = output_something(outputs=[of])
    output_file_future = app_future.outputs[0]

    app_future.result()
    output_file_future.result()

    assert zipfile.is_zipfile(zip_path)

    with zipfile.ZipFile(zip_path) as z:
        assert file_base in z.namelist()
        assert len(z.namelist()) == 1
        with z.open(file_base) as f:
            assert f.readlines() == [b'0\n']


@pytest.mark.local
def test_zip_out_multi(tmpd_cwd):
    # tests multiple files, multiple zip files and multiple
    # sub-paths

    zip_path_1 = tmpd_cwd / "container1.zip"
    zip_path_2 = tmpd_cwd / "container2.zip"

    relative_file_path_1 = "a/b/c/data.txt"
    relative_file_path_2 = "something.txt"
    relative_file_path_3 = "a/d/other.txt"
    of1 = File(f"zip:{zip_path_1 / relative_file_path_1}")
    of2 = File(f"zip:{zip_path_1 / relative_file_path_2}")
    of3 = File(f"zip:{zip_path_2 / relative_file_path_3}")

    app_future = output_something(outputs=[of1, of2, of3])

    for f in app_future.outputs:
        f.result()

    app_future.result()

    assert zipfile.is_zipfile(zip_path_1)

    with zipfile.ZipFile(zip_path_1) as z:
        assert relative_file_path_1 in z.namelist()
        assert relative_file_path_2 in z.namelist()
        assert len(z.namelist()) == 2
        with z.open(relative_file_path_1) as f:
            assert f.readlines() == [b'0\n']
        with z.open(relative_file_path_2) as f:
            assert f.readlines() == [b'1\n']

    assert zipfile.is_zipfile(zip_path_2)

    with zipfile.ZipFile(zip_path_2) as z:
        assert relative_file_path_3 in z.namelist()
        assert len(z.namelist()) == 1
        with z.open(relative_file_path_3) as f:
            assert f.readlines() == [b'2\n']


@pytest.mark.local
def test_zip_bad_authority(tmpd_cwd):
    # tests that there's an exception when staging a ZIP url with an authority
    # section specified, rather than silently ignoring it. This simulates a
    # user who misunderstands what that piece of what a zip: URL means.

    zip_path = tmpd_cwd / "container.zip"
    file_base = "data.txt"
    of = File(f"zip://someauthority/{zip_path / file_base}")

    with pytest.raises(ZipAuthorityError):
        output_something(outputs=[of])



from enum import IntEnum


class States(IntEnum):
    """Enumerates the states a parsl task may be in.

    These states occur inside the task record for a task inside
    a `DataFlowKernel` and in the monitoring database.

    In a single successful task execution, tasks will progress in this
    sequence:

    pending -> launched -> running -> running_ended -> exec_done

    Other states represent deviations from this path, either due to
    failures, or to deliberate changes to how tasks are executed (for
    example due to join_app, or memoization).


    All tasks should end up in one of the states listed in `FINAL_STATES`.
    """

    pending = 0
    """Task is known to parsl but cannot run yet. Usually, a task cannot
    run because it is waiting for dependency tasks to complete.
    """

    running = 2
    """Task is running on a resource. This state is special - a DFK task
    record never goes to States.running state; but the monitoring database
    may represent a task in this state based on non-DFK information received
    from monitor_wrapper."""

    exec_done = 3
    """Task has been executed successfully."""

    failed = 4
    """Task has failed and no more attempts will be made to run it."""

    dep_fail = 5
    """Dependencies of this task failed, so it is marked as failed without
    even an attempt to launch it."""

    launched = 7
    """Task has been passed to a `ParslExecutor` for execution."""

    fail_retryable = 8
    """Task has failed, but can be retried"""

    memo_done = 9
    """Task was found in the memoization table, so it is marked as done
    without even an attempt to launch it."""

    joining = 10
    """Task is a join_app, joining on internal tasks. The task has run its
    own Python code, and is now waiting on other tasks before it can make
    further progress (to a done/failed state)."""

    running_ended = 11
    """Like States.running, this state is also not observed by the DFK,
    but instead only by monitoring. This state does not record
    anything about task success or failure, merely that the wrapper
    ran long enough to record it as finished."""

    def __str__(self) -> str:
        return self.__class__.__name__ + "." + self.name


FINAL_FAILURE_STATES = {States.failed, States.dep_fail}
"""States which are final and which indicate a failure. This must
be a subset of FINAL_STATES"""

FINAL_STATES = {States.exec_done, States.memo_done, *FINAL_FAILURE_STATES}
"""States from which we will never move to another state, because the job has
either definitively completed or failed."""



from typing import Optional


class ParslTaskToVine:
    """ Helper class to transform a Parsl function into a TaskVine task."""

    def __init__(self,
                 executor_id: int,                 # executor id of Parsl function
                 exec_mode: str,                   # execution mode of function, out of {regular, python, serverless}
                 category: str,                    # category of Parsl function
                 input_files: list,                # list of input files to this function
                 output_files: list,               # list of output files to this function
                 map_file: Optional[str],          # pickled file containing mapping of local to remote names of files
                 function_file: Optional[str],     # pickled file containing the function information
                 argument_file: Optional[str],     # pickled file containing the arguments to the function call
                 result_file: Optional[str],       # path to the pickled result object of the function execution
                 cores: Optional[float],           # number of cores to allocate
                 memory: Optional[int],            # amount of memory in MBs to allocate
                 disk: Optional[int],              # amount of disk in MBs to allocate
                 gpus: Optional[float],            # number of gpus to allocate
                 priority: Optional[float],        # priority of function, the higher the more priority
                 running_time_min: Optional[int],  # minimum amount of time for the function to run on a worker
                 env_pkg: Optional[str],           # path to a poncho environment tarball
                 ):
        self.executor_id = executor_id
        self.exec_mode = exec_mode
        self.category = category
        self.map_file = map_file
        self.function_file = function_file
        self.argument_file = argument_file
        self.result_file = result_file
        self.input_files = input_files
        self.output_files = output_files
        self.cores = cores
        self.memory = memory
        self.disk = disk
        self.gpus = gpus
        self.priority = priority
        self.running_time_min = running_time_min
        self.env_pkg = env_pkg


class VineTaskToParsl:
    """
    Support structure to communicate final status of TaskVine tasks to Parsl.
    result_file is only valid if result_received is True.
    Reason and status are only valid if result_received is False.
    """
    def __init__(self,
                 executor_id: int,          # executor id of task
                 result_received: bool,     # whether result is received or not
                 result_file: Optional[str],    # path to file that contains the serialized result object
                 reason: Optional[str],     # string describing why execution fails
                 status: Optional[int]      # exit code of execution of task
                 ):
        self.executor_id = executor_id
        self.result_received = result_received
        self.result_file = result_file
        self.reason = reason
        self.status = status


class ParslFileToVine:
    """
    Support structure to report Parsl filenames to TaskVine.
    parsl_name is the local_name or filepath attribute of a Parsl file object.
    stage tells whether the file should be copied by TaskVine to the workers.
    cache tells whether the file should be cached at workers. Only valid if stage == True
    """
    def __init__(self,
                 parsl_name: str,   # name of file
                 stage: bool,       # whether TaskVine should know about this file
                 cache: bool        # whether TaskVine should cache this file
                 ):
        self.parsl_name = parsl_name
        self.stage = stage
        self.cache = cache


def run_parsl_function(map_file, function_file, argument_file, result_file):
    """
    Wrapper function to deploy with FunctionCall as serverless tasks.
    """
    from parsl.executors.taskvine.exec_parsl_function import run
    run(map_file, function_file, argument_file, result_file)



import os

import pytest

from parsl.app.app import bash_app, python_app
from parsl.data_provider.files import File


@bash_app
def echo(message, outputs=[]):
    return 'echo {m} &> {o}'.format(m=message, o=outputs[0])


@python_app
def cat(inputs=[]):
    with open(inputs[0].filepath) as f:
        return f.readlines()


@pytest.mark.staging_required
def test_slides():
    """Testing code snippet from slides """

    if os.path.exists('hello1.txt'):
        os.remove('hello1.txt')

    hello = echo("Hello World!", outputs=[File('hello1.txt')])

    message = cat(inputs=[hello.outputs[0]])

    # Waits. This need not be in the slides.
    print(hello.result())
    print(message.result())



from typing import Any

import pytest

import parsl
from parsl.serialize.facade import methods_for_code
from parsl.tests.configs.htex_local import fresh_config as local_config


@parsl.python_app
def f(x):
    return x + 1


@pytest.mark.local
def test_caching() -> None:
    # for future serializer devs: if this is failing because you added another
    # code serializer, you'll also probably need to re-think what is being tested
    # about serialization caching here.
    assert len(methods_for_code) == 1

    serializer = methods_for_code[b'C2']

    # force type to Any here because a serializer method coming from
    # methods_for_code doesn't statically have any cache management
    # methods on itself such as cache_clear or cache_info.
    serialize_method: Any = serializer.serialize

    serialize_method.cache_clear()

    assert serialize_method.cache_info().hits == 0
    assert serialize_method.cache_info().misses == 0
    assert serialize_method.cache_info().currsize == 0

    assert f(7).result() == 8

    # the code serializer cache should now contain only a (probably wrapped) f ...
    assert serialize_method.cache_info().currsize == 1

    # ... which was not already in the cache.
    assert serialize_method.cache_info().misses == 1
    assert serialize_method.cache_info().hits == 0

    assert f(100).result() == 101

    # this time round, we should have got a single cache hit...
    assert serialize_method.cache_info().hits == 1
    assert serialize_method.cache_info().misses == 1
    assert serialize_method.cache_info().currsize == 1

    assert f(200).result() == 201

    # this time round, we should have got another single cache hit...
    assert serialize_method.cache_info().hits == 2
    assert serialize_method.cache_info().misses == 1
    assert serialize_method.cache_info().currsize == 1



import pytest

from parsl.serialize import deserialize, serialize
from parsl.serialize.concretes import DillSerializer, PickleSerializer


@pytest.mark.local
def test_serialize():
    assert deserialize(serialize(1)) == 1


@pytest.mark.local
def test_pickle_wrapper():
    s = PickleSerializer()
    d = s.serialize(1)
    assert isinstance(d, bytes)
    assert s.deserialize(d) == 1


@pytest.mark.local
def test_dill_wrapper():
    s = DillSerializer()
    d = s.serialize(1)
    assert isinstance(d, bytes)
    assert s.deserialize(d) == 1



import logging
import pickle
from multiprocessing.queues import Queue

from parsl.monitoring.radios.base import (
    MonitoringRadioReceiver,
    MonitoringRadioSender,
    RadioConfig,
)
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)


class HTEXRadio(RadioConfig, RepresentationMixin):
    def create_sender(self) -> MonitoringRadioSender:
        return HTEXRadioSender()

    def create_receiver(self, *, run_dir: str, resource_msgs: Queue) -> MonitoringRadioReceiver:
        return HTEXRadioReceiver()


class HTEXRadioSender(MonitoringRadioSender):

    def __init__(self) -> None:
        # there is nothing to initialize
        pass

    def send(self, message: object) -> None:
        """ Sends a message to the UDP receiver

        Parameter
        ---------

        message: object
            Arbitrary pickle-able object that is to be sent

        Returns:
            None
        """

        import parsl.executors.high_throughput.monitoring_info

        result_queue = parsl.executors.high_throughput.monitoring_info.result_queue

        # this message needs to go in the result queue tagged so that it is treated
        # i) as a monitoring message by the interchange, and then further more treated
        # as a RESOURCE_INFO message when received by monitoring (rather than a NODE_INFO
        # which is the implicit default for messages from the interchange)

        # for the interchange, the outer wrapper, this needs to be a dict:

        interchange_msg = {
            'type': 'monitoring',
            'payload': message
        }

        if result_queue:
            result_queue.put(pickle.dumps(interchange_msg))
        else:
            logger.error("result_queue is uninitialized - cannot put monitoring message")


class HTEXRadioReceiver(MonitoringRadioReceiver):
    def shutdown(self) -> None:
        # there is nothing to shut down
        pass



import os
from pathlib import Path

import pytest

import parsl
from parsl import python_app
from parsl.config import Config
from parsl.dataflow.memoization import BasicMemoizer


@python_app(cache=True)
def uuid_app():
    import uuid
    return uuid.uuid4()


@pytest.mark.local
def test_manual_checkpoint() -> None:
    """Launch a single app, write manually, check file exists.
    """
    memoizer = BasicMemoizer(checkpoint_mode="manual")

    with parsl.load(Config(memoizer=memoizer)):
        uuid_app().result()

        cpt_dir = Path(parsl.dfk().run_dir) / 'checkpoint'

        cptpath = cpt_dir / 'tasks.pkl'

        assert not os.path.exists(cptpath), f"Tasks checkpoint should not exist yet: {cptpath}"
        memoizer.checkpoint()
        assert os.path.exists(cptpath), f"Tasks checkpoint should exist now: {cptpath}"



import random
import threading

import pytest

import parsl
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import LocalProvider

# we need some blocks, but it doesn't matter too much how many, as long
# as they can all start up and get registered within the try_assert
# timeout later on.
BLOCK_COUNT = 3

# the try_assert timeout for the above number of blocks to get started
PERMITTED_STARTUP_TIME_S = 30


class AccumulatingLocalProvider(LocalProvider):
    def __init__(self, *args, **kwargs):
        # Use a list for submitted job IDs because if there are multiple
        # submissions returning the same job ID, this test should count
        # those...
        self.submit_job_ids = []

        # ... but there's no requirement, I think, that cancel must be called
        # only once per job id. What matters here is that each job ID is
        # cancelled at least once.
        self.cancel_job_ids = set()

        super().__init__(*args, **kwargs)

    def submit(self, *args, **kwargs):
        job_id = super().submit(*args, **kwargs)
        self.submit_job_ids.append(job_id)
        return job_id

    def cancel(self, job_ids):
        self.cancel_job_ids.update(job_ids)
        return super().cancel(job_ids)


@pytest.mark.local
def test_shutdown_scalein_blocks(tmpd_cwd, try_assert):
    """
    This test scales up several blocks, and then checks that they are all
    scaled in at DFK shutdown.
    """
    accumulating_provider = AccumulatingLocalProvider(
        init_blocks=BLOCK_COUNT,
        min_blocks=0,
        max_blocks=0,
        launcher=SimpleLauncher(),
    )

    htex = HighThroughputExecutor(
               label="htex_local",
               cores_per_worker=1,
               provider=accumulating_provider
           )

    config = Config(
        executors=[htex],
        strategy='none',
        strategy_period=0.1,
        run_dir=str(tmpd_cwd)
    )

    with parsl.load(config):
        # this will wait for everything to be scaled out fully
        try_assert(lambda: len(htex.connected_managers()) == BLOCK_COUNT, timeout_ms=PERMITTED_STARTUP_TIME_S * 1000)

    assert len(accumulating_provider.submit_job_ids) == BLOCK_COUNT, f"Exactly {BLOCK_COUNT} blocks should have been launched"
    assert len(accumulating_provider.cancel_job_ids) == BLOCK_COUNT, f"Exactly {BLOCK_COUNT} blocks should have been scaled in"



import time

import pytest

import parsl
from parsl.app.app import bash_app, python_app
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.executors.high_throughput.manager_selector import (
    BlockIdManagerSelector,
    ManagerSelector,
)
from parsl.launchers import WrappedLauncher
from parsl.providers import LocalProvider
from parsl.usage_tracking.levels import LEVEL_1

BLOCK_COUNT = 2


@parsl.python_app
def get_worker_pid():
    import os
    return os.environ.get('PARSL_WORKER_BLOCK_ID')


@pytest.mark.local
def test_block_id_selection(try_assert):
    htex = HighThroughputExecutor(
        label="htex_local",
        max_workers_per_node=1,
        manager_selector=BlockIdManagerSelector(),
        provider=LocalProvider(
            init_blocks=BLOCK_COUNT,
            max_blocks=BLOCK_COUNT,
            min_blocks=BLOCK_COUNT,
        ),
    )

    config = Config(
        executors=[htex],
        usage_tracking=LEVEL_1,
    )

    with parsl.load(config):
        blockids = []
        try_assert(lambda: len(htex.connected_managers()) == BLOCK_COUNT, timeout_ms=20000)
        for i in range(10):
            future = get_worker_pid()
            blockids.append(future.result())

        assert all(blockid == "1" for blockid in blockids)
